# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2018 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that connects to one crazyflie (check the address at the top
and update it to your crazyflie address) and uses the high level commander
to send setpoints and trajectory to fly a figure 8.

This example is intended to work with any positioning system (including LPS).
It aims at documenting how to set the Crazyflie in position control mode
and how to send setpoints using the high level commander.
"""
import sys
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.high_level_commander import HighLevelCommander
from cflib.crazyflie.mem import CompressedSegment
from cflib.crazyflie.mem import CompressedStart
from cflib.crazyflie.mem import MemoryElement
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper
from cflib.utils.reset_estimator import reset_estimator

# URI to the Crazyflie to connect to
uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# The trajectory to fly
a = 0.9
b = 0.5
c = 0.5
trajectory = [
    CompressedStart(0.0, 0.0, 0.0, 0.0),
    CompressedSegment(2.0, [0.0, 1.0, 1.0], [0.0, a, 0.0], [], []),
    CompressedSegment(2.0, [1.0, b, 0.0], [-a, -c, 0.0], [], []),
    CompressedSegment(2.0, [-b, -1.0, -1.0], [c, a, 0.0], [], []),
    CompressedSegment(2.0, [-1.0, 0.0, 0.0], [-a, 0.0, 0.0], [], []),
]


def activate_mellinger_controller(cf):
    cf.param.set_value('stabilizer.controller', '2')


def upload_trajectory(cf, trajectory_id, trajectory):
    trajectory_mem = cf.mem.get_mems(MemoryElement.TYPE_TRAJ)[0]

    trajectory_mem.trajectory = trajectory

    upload_result = trajectory_mem.write_data_sync()
    if not upload_result:
        print('Upload failed, aborting!')
        sys.exit(1)
    cf.high_level_commander.define_trajectory(
        trajectory_id,
        0,
        len(trajectory),
        type=HighLevelCommander.TRAJECTORY_TYPE_POLY4D_COMPRESSED)

    total_duration = 0
    # Skip the start element
    for segment in trajectory[1:]:
        total_duration += segment.duration

    return total_duration


def run_sequence(cf, trajectory_id, duration):
    commander = cf.high_level_commander

    # Arm the Crazyflie
    cf.platform.send_arming_request(True)
    time.sleep(1.0)

    commander.takeoff(1.0, 2.0)
    time.sleep(3.0)
    relative = True
    commander.start_trajectory(trajectory_id, 1.0, relative)
    time.sleep(duration)
    commander.land(0.0, 2.0)
    time.sleep(2)
    commander.stop()


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
        cf = scf.cf
        trajectory_id = 1

        # activate_mellinger_controller(cf)
        duration = upload_trajectory(cf, trajectory_id, trajectory)
        print('The sequence is {:.1f} seconds long'.format(duration))
        reset_estimator(cf)
        run_sequence(cf, trajectory_id, duration)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2024 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import unittest
from unittest.mock import ANY
from unittest.mock import mock_open
from unittest.mock import patch

import yaml

from cflib.localization import ParamFileManager


class TestParamFileManager(unittest.TestCase):
    def setUp(self):
        self.data = {
            'type': 'persistent_param_state',
            'version': '1',
        }

    @patch('yaml.safe_load')
    def test_that_read_open_correct_file(self, mock_yaml_load):
        # Fixture
        mock_yaml_load.return_value = self.data
        file_name = 'some/name.yaml'

        # Test
        with patch('builtins.open', new_callable=mock_open()) as mock_file:
            ParamFileManager.read(file_name)

        # Assert
        mock_file.assert_called_with(file_name, 'r')

    @patch('yaml.safe_load')
    def test_that_missing_file_type_raises(self, mock_yaml_load):
        # Fixture
        self.data.pop('type')
        mock_yaml_load.return_value = self.data

        # Test
        # Assert
        with self.assertRaises(Exception):
            with patch('builtins.open', new_callable=mock_open()):
                ParamFileManager.read('some/name.yaml')

    @patch('yaml.safe_load')
    def test_that_wrong_file_type_raises(self, mock_yaml_load):
        # Fixture
        self.data['type'] = 'wrong_type'
        mock_yaml_load.return_value = self.data

        # Test
        # Assert
        with self.assertRaises(Exception):
            with patch('builtins.open', new_callable=mock_open()):
                ParamFileManager.read('some/name.yaml')

    @patch('yaml.safe_load')
    def test_that_missing_version_raises(self, mock_yaml_load):

        # Fixture
        self.data.pop('version')
        mock_yaml_load.return_value = self.data

        # Test
        # Assert
        with self.assertRaises(Exception):
            with patch('builtins.open', new_callable=mock_open()):
                ParamFileManager.read('some/name.yaml')

    @patch('yaml.safe_load')
    def test_that_wrong_version_raises(self, mock_yaml_load):
        # Fixture
        self.data['version'] = 'wrong_version'
        mock_yaml_load.return_value = self.data

        # Test
        # Assert
        with self.assertRaises(Exception):
            with patch('builtins.open', new_callable=mock_open()):
                ParamFileManager.read('some/name.yaml')

    @patch('yaml.safe_load')
    def test_that_no_data_returns_empty_default_data(self, mock_yaml_load):
        # Fixture
        mock_yaml_load.return_value = self.data

        # Test
        with patch('builtins.open', new_callable=mock_open()):
            actual_params = ParamFileManager.read('some/name.yaml')

        # Assert
        self.assertEqual(0, len(actual_params))

    @patch('yaml.dump')
    def test_file_end_to_end_write_read(self, mock_yaml_dump):
        # Fixture
        fixture_file = 'test/localization/fixtures/parameters.yaml'

        file = open(fixture_file, 'r')
        expected = yaml.safe_load(file)
        file.close()

        # Test
        params = ParamFileManager.read(fixture_file)
        with patch('builtins.open', new_callable=mock_open()):
            ParamFileManager.write('some/name.yaml', params=params)

            # Assert
            mock_yaml_dump.assert_called_with(expected, ANY)

    @patch('yaml.dump')
    def test_file_write_to_correct_file(self, mock_yaml_dump):
        # Fixture
        file_name = 'some/name.yaml'

        # Test
        with patch('builtins.open', new_callable=mock_open()) as mock_file:
            ParamFileManager.write(file_name)

            # Assert
            mock_file.assert_called_with(file_name, 'w')



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017-2018 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example of an asynchronized swarm choreography using the motion
commander.

The swarm takes off and flies an asynchronous choreography before landing.
All movements are relative to the starting position.
During the flight, the position of each Crazyflie is printed.

This example is intended to work with any kind  of location system, it has
been tested with the flow deck v2 and the lighthouse positioning system.
Not using an absolute positioning system makes every Crazyflie start its
positioning printing with (0,0,0) as its initial position.

This example aims at documenting how to use the motion commander together
with the Swarm class to achieve asynchronized sequences.
"""
import time

import cflib.crtp
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm
from cflib.positioning.motion_commander import MotionCommander

# Change uris according to your setup
# URIs in a swarm using the same radio must also be on the same channel
URI1 = 'radio://0/80/2M/E7E7E7E7E7'
URI2 = 'radio://0/80/2M/E7E7E7E7E8'

DEFAULT_HEIGHT = 0.5
DEFAULT_VELOCITY = 0.2
pos1 = [0, 0, 0]
pos2 = [0, 0, 0]

# List of URIs
uris = {
    URI1,
    URI2,
}


def wait_for_param_download(scf):
    while not scf.cf.param.is_updated:
        time.sleep(1.0)
    print('Parameters downloaded for', scf.cf.link_uri)


def arm(scf):
    scf.cf.platform.send_arming_request(True)
    time.sleep(1.0)


def position_callback(uri, data):
    if uri == URI1:
        pos1[0] = data['stateEstimate.x']
        pos1[1] = data['stateEstimate.y']
        pos1[2] = data['stateEstimate.z']
        print(f'Uri1 position: x={pos1[0]}, y={pos1[1]}, z={pos1[2]}')
    elif uri == URI2:
        pos2[0] = data['stateEstimate.x']
        pos2[1] = data['stateEstimate.y']
        pos2[2] = data['stateEstimate.z']
        print(f'Uri2 position: x={pos2[0]}, y={pos2[1]}, z={pos2[2]}')


def start_position_printing(scf):
    log_conf1 = LogConfig(name='Position', period_in_ms=500)
    log_conf1.add_variable('stateEstimate.x', 'float')
    log_conf1.add_variable('stateEstimate.y', 'float')
    log_conf1.add_variable('stateEstimate.z', 'float')
    scf.cf.log.add_config(log_conf1)
    log_conf1.data_received_cb.add_callback(lambda _timestamp, data, _logconf: position_callback(scf.cf.link_uri, data))
    log_conf1.start()


def async_flight(scf):
    with MotionCommander(scf, default_height=DEFAULT_HEIGHT) as mc:
        time.sleep(1)

        start_time = time.time()
        end_time = time.time() + 12

        while time.time() < end_time:

            if scf.__dict__['_link_uri'] == URI1:
                if time.time() - start_time < 5:
                    mc.start_up(DEFAULT_VELOCITY)
                elif time.time() - start_time < 7:
                    mc.stop()
                elif time.time() - start_time < 12:
                    mc.start_down(DEFAULT_VELOCITY)
                else:
                    mc.stop()

            elif scf.__dict__['_link_uri'] == URI2:
                if time.time() - start_time < 2:
                    mc.start_left(DEFAULT_VELOCITY)
                elif time.time() - start_time < 4:
                    mc.start_right(DEFAULT_VELOCITY)
                elif time.time() - start_time < 6:
                    mc.start_left(DEFAULT_VELOCITY)
                elif time.time() - start_time < 8:
                    mc.start_right(DEFAULT_VELOCITY)
                elif time.time() - start_time < 10:
                    mc.start_left(DEFAULT_VELOCITY)
                elif time.time() - start_time < 12:
                    mc.start_right(DEFAULT_VELOCITY)
                else:
                    mc.stop()

            time.sleep(0.01)
        mc.land()


if __name__ == '__main__':
    # logging.basicConfig(level=logging.DEBUG)
    cflib.crtp.init_drivers()

    factory = CachedCfFactory(rw_cache='./cache')
    with Swarm(uris, factory=factory) as swarm:

        swarm.reset_estimators()

        print('Waiting for parameters to be downloaded...')
        swarm.parallel_safe(wait_for_param_download)

        time.sleep(1)

        swarm.parallel_safe(start_position_printing)
        time.sleep(0.1)

        swarm.parallel_safe(arm)
        time.sleep(0.5)

        swarm.parallel_safe(async_flight)
        time.sleep(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Bootloading utilities for the Crazyflie.
"""
import json
import logging
import sys
import time
import zipfile
from collections import namedtuple
from typing import Callable
from typing import List
from typing import NoReturn
from typing import Optional
from typing import Tuple

from packaging.version import Version

from .boottypes import BootVersion
from .boottypes import TargetTypes
from .cloader import Cloader
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.mem import deck_memory
from cflib.crazyflie.mem import MemoryElement
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils.power_switch import PowerSwitch

logger = logging.getLogger(__name__)

__author__ = 'Bitcraze AB'
__all__ = ['Bootloader']

Target = namedtuple('Target', ['platform', 'target', 'type', 'provides', 'requires'])
FlashArtifact = namedtuple('FlashArtifact', ['content', 'target', 'release'])


class Bootloader:
    """Bootloader utility for the Crazyflie"""

    def __init__(self, clink=None):
        """Init the communication class by starting to communicate with the
        link given. clink is the link address used after resetting to the
        bootloader.

        The device is actually considered in firmware mode.
        """
        self.clink = clink
        self.in_loader = False

        self.page_size = 0
        self.buffer_pages = 0
        self.flash_pages = 0
        self.start_page = 0
        self.cpuid = 'N/A'
        self.error_code = 0
        self.protocol_version = 0

        self.warm_booted = False

        # Outgoing callbacks for progress and flash termination
        self.progress_cb = None  # type: Optional[Callable[[str, int], None]]
        self.error_cb = None  # type: Optional[Callable[[str], None]]
        self.terminate_flashing_cb = None  # type: Optional[Callable[[], bool]]

        self._boot_plat = None

        self._cload = Cloader(clink,
                              info_cb=None,
                              in_boot_cb=None)

    def start_bootloader(self, warm_boot=False, cf=None):
        self.warm_booted = warm_boot

        if warm_boot:
            if cf is not None and cf.link:
                cf.close_link()
            self._cload.open_bootloader_uri(self.clink)
            started = self._cload.reset_to_bootloader(TargetTypes.NRF51)
            if started:
                started = self._cload.check_link_and_get_info()
        else:
            if not self._cload.link:
                uri = self._cload.scan_for_bootloader()

                # Workaround for libusb on Windows (open/close too fast)
                time.sleep(1)

                if uri:
                    self._cload.open_bootloader_uri(uri)
                    started = self._cload.check_link_and_get_info()
                else:
                    started = False
            else:
                started = True
        if started:
            self.protocol_version = self._cload.protocol_version

            if (self.protocol_version == BootVersion.CF1_PROTO_VER_0 or
                    self.protocol_version == BootVersion.CF1_PROTO_VER_1):
                # Nothing more to do
                pass
            elif self.protocol_version == BootVersion.CF2_PROTO_VER:
                self._cload.request_info_update(TargetTypes.NRF51)
            else:
                print('Bootloader protocol 0x{:X} not '
                      'supported!'.format(self.protocol_version))

        return started

    def get_target(self, target_id):
        return self._cload.request_info_update(target_id)

    def _get_current_nrf51_sd_version(self):
        if self._cload.targets[TargetTypes.NRF51].start_page == 88:
            return 'sd-s110'
        elif self._cload.targets[TargetTypes.NRF51].start_page == 108:
            return 'sd-s130'
        else:
            raise Exception('Unknown soft device running on nRF51')

    def _get_required_nrf51_sd_version(self, flash_artifacts: List[FlashArtifact]):
        required_nrf_sd_version = None
        for a in flash_artifacts:
            if a.target.target == 'nrf51':
                for r in a.target.requires:
                    if required_nrf_sd_version is None:
                        required_nrf_sd_version = r
                    if required_nrf_sd_version != r:
                        raise Exception('Cannot flash nRF51, conflicting requirements: {} and {}'.format(
                            required_nrf_sd_version, r))

        return required_nrf_sd_version

    def _get_provided_nrf51_sd_version(self, flash_artifacts: List[FlashArtifact]):
        provided_nrf_sd_version = None
        for a in flash_artifacts:
            if a.target.target == 'nrf51':
                for r in a.target.provides:
                    if provided_nrf_sd_version is None:
                        provided_nrf_sd_version = r
                    if provided_nrf_sd_version != r:
                        raise Exception('Cannot flash nRF51, conflicting requirements: {} and {}'.format(
                            provided_nrf_sd_version, r))

        return provided_nrf_sd_version

    def _get_provided_nrf51_bl_version(self, flash_artifacts: List[FlashArtifact]):
        provided_nrf_bl_version = None
        for a in flash_artifacts:
            if a.target.target == 'nrf51' and a.target.type == 'bootloader+softdevice':
                if provided_nrf_bl_version is None:
                    provided_nrf_bl_version = a.release
                else:
                    raise Exception('One and only one bootloader+softdevice in zip file supported')

        return provided_nrf_bl_version

    def _get_boot_delay(self, cf: Optional[Crazyflie] = None):
        """Determines the boot delay based on AI-deck presence or failure to connect."""
        try:
            # First try using an existing link
            if cf.link:
                return 5.0 if self._has_ai_deck(cf) else 0.0
        except Exception:
            # Failed to connect using existing link, try opening a new link
            pass

        try:
            with SyncCrazyflie(cf.uri, cf=Crazyflie()) as scf:
                return 5.0 if self._has_ai_deck(scf.cf) else 0.0
        except Exception:
            # If we fail to connect using SyncCrazyflie, assume the Crazyflie may be in bootloader mode
            return 5.0

    def _has_ai_deck(self, cf):
        deck_mems = cf.mem.get_mems(MemoryElement.TYPE_DECK_MEMORY)
        if not deck_mems:
            raise RuntimeError('Failed to read memory: No deck memory found')

        mgr = deck_memory.SyncDeckMemoryManager(deck_mems[0])
        try:
            decks = mgr.query_decks()
        except RuntimeError as e:
            if self.progress_cb:
                message = f'Failed to read decks: {str(e)}'
                self.progress_cb(message, 0)
                print(message)
                time.sleep(2)
                raise RuntimeError(message)

        return any(deck.name in ['bcAI:gap8', 'bcAI:esp'] for deck in decks.values())

    def flash(self, filename: str, targets: List[Target], cf=None, enable_console_log: Optional[bool] = False,
              boot_delay: Optional[float] = 0.0):
        # Separate flash targets from decks
        platform = self._get_platform_id()
        flash_targets = [t for t in targets if t.platform == platform]
        deck_targets = [t for t in targets if t.platform == 'deck']

        # Fetch artifacts from source file
        artifacts = self._get_flash_artifacts_from_zip(filename)
        if len(artifacts) == 0:
            if len(targets) == 1:
                content = open(filename, 'br').read()
                artifacts = [FlashArtifact(content, targets[0], None)]
            else:
                raise (Exception('Cannot flash a .bin to more than one target!'))

        # Separate artifacts for flash and decks
        flash_artifacts = [a for a in artifacts if a.target.platform == platform]
        deck_artifacts = [a for a in artifacts if a.target.platform == 'deck']

        # Handle the special case of flashing soft device and bootloader for nRF51
        current_nrf_sd_version = self._get_current_nrf51_sd_version()
        required_nrf_sd_version = self._get_required_nrf51_sd_version(flash_artifacts)
        provided_nrf_sd_version = self._get_provided_nrf51_sd_version(flash_artifacts)
        update_contains_nrf_sd = any(x.target.type == 'bootloader+softdevice' for x in flash_artifacts)
        current_nrf_bl_version = None
        if self._cload.targets[TargetTypes.NRF51].version is not None:
            current_nrf_bl_version = Version(str(self._cload.targets[TargetTypes.NRF51].version))
        provided_nrf_bl_version = None
        if self._get_provided_nrf51_bl_version(flash_artifacts) is not None:
            provided_nrf_bl_version = Version(self._get_provided_nrf51_bl_version(flash_artifacts))

        print('nRF51 has: {} and requires {} and upgrade provides {}. Current bootloader version is [{}] and upgrade '
              'provides [{}]'.format(
                  current_nrf_sd_version, required_nrf_sd_version, provided_nrf_sd_version,
                  current_nrf_bl_version, provided_nrf_bl_version)
              )

        if required_nrf_sd_version is not None and \
                current_nrf_sd_version != required_nrf_sd_version and \
                provided_nrf_sd_version != required_nrf_sd_version:
            raise Exception('Cannot flash nRF51: We have sd {}, need {} and have a zip with {}'.format(
                current_nrf_sd_version, required_nrf_sd_version, provided_nrf_sd_version))

        # To avoid always flashing the bootloader and soft device (these might never change again) first check
        # if we really need to. The original version of the bootloader is reported as None.
        should_flash_nrf_sd = True
        if current_nrf_sd_version == required_nrf_sd_version and current_nrf_bl_version == provided_nrf_bl_version:
            should_flash_nrf_sd = False
        elif provided_nrf_sd_version is None and not update_contains_nrf_sd:
            should_flash_nrf_sd = False

        if should_flash_nrf_sd:
            print('Should flash nRF soft device')
            rf51_sdbl_list = [x for x in flash_artifacts if x.target.type == 'bootloader+softdevice']
            if len(rf51_sdbl_list) != 1:
                raise Exception('Only support for one and only one bootloader+softdevice in zip file')
            nrf51_sdbl = rf51_sdbl_list[0]

            nrf_info = self._cload.targets[TargetTypes.NRF51]
            # During bootloader update part of the firmware will be erased. If we're
            # only flashing the bootloader and no firmware we should make sure the bootloader
            # stays in bootloader mode and doesn't try to start the broken firmware, this is
            # done by erasing the first page of the firmware.
            self._internal_flash(FlashArtifact([0xFF] * nrf_info.page_size, nrf51_sdbl.target, None))
            page = nrf_info.flash_pages - (len(nrf51_sdbl.content) // nrf_info.page_size)
            self._internal_flash(artifact=nrf51_sdbl, page_override=page)

            self._cload.reset_to_bootloader(TargetTypes.NRF51)
            uri = self._cload.link.uri
            self._cload.close()
            print('Closing bootloader link and reconnecting to new bootloader (' + uri + ')')
            self._cload = Cloader(uri,
                                  info_cb=None,
                                  in_boot_cb=None)
            self._cload.open_bootloader_uri(uri)
            print('Reconnected to new bootloader')
            self._cload.check_link_and_get_info()
            self._cload.request_info_update(TargetTypes.NRF51)
        else:
            print('No need to flash nRF soft device')

        # Remove the softdevice+bootloader from the list of artifacts to flash
        flash_artifacts = [a for a in flash_artifacts if a.target.type !=
                           'bootloader+softdevice']  # Also filter for nRF51 here?

        # Flash the MCU flash
        if len(targets) == 0 or len(flash_targets) > 0:
            self._flash_flash(flash_artifacts, flash_targets)

        # Flash the decks
        deck_update_msg = 'Deck update skipped.'
        if len(targets) == 0 or len(deck_targets) > 0:
            # only in warm boot
            if self.warm_booted:
                if self.progress_cb:
                    self.progress_cb('Restarting firmware to update decks.', int(0))

                # Reset to firmware mode
                self.reset_to_firmware(boot_delay=boot_delay)
                self.close()
                time.sleep(2)

                # Flash all decks and reboot after each deck
                current_index = 0
                while current_index != -1:
                    current_index = self._flash_deck_incrementally(deck_artifacts, deck_targets, current_index,
                                                                   enable_console_log=enable_console_log,
                                                                   boot_delay=boot_delay)
                    if self.progress_cb:
                        self.progress_cb('Deck updated! Restarting...', int(100))
                    if current_index != -1:
                        PowerSwitch(self.clink).reboot_to_fw()
                        time.sleep(2.0 + boot_delay)

                # Put the crazyflie back in Bootloader mode to exit the function in the same state we entered it
                self.start_bootloader(warm_boot=True, cf=cf)

                deck_update_msg = 'Deck update complete.'
            else:
                print('Skipping updating deck on coldboot')
                deck_update_msg = 'Deck update skipped in ColdBoot mode.'

        if self.progress_cb:
            self.progress_cb(
                f'({len(flash_artifacts)}/{len(flash_artifacts)}) Flashing done! {deck_update_msg}',
                int(100))
        else:
            print('')

    def flash_full(self, cf: Optional[Crazyflie] = None,
                   filename: Optional[str] = None,
                   warm: bool = True,
                   targets: Optional[Tuple[str, ...]] = None,
                   info_cb: Optional[Callable[[int, TargetTypes], NoReturn]] = None,
                   progress_cb: Optional[Callable[[str, int], NoReturn]] = None,
                   terminate_flash_cb: Optional[Callable[[], bool]] = None,
                   enable_console_log: Optional[bool] = False):
        """
        Flash .zip or bin .file to list of targets.
        Reset to firmware when done.
        """
        # Get the required boot delay
        boot_delay = self._get_boot_delay(cf=cf)

        if progress_cb is not None:
            self.progress_cb = progress_cb
        if terminate_flash_cb is not None:
            self.terminate_flashing_cb = terminate_flash_cb

        if not self.start_bootloader(warm_boot=warm, cf=cf):
            raise Exception('Could not connect to bootloader')

        if info_cb is not None:
            connected = (self.get_target(TargetTypes.STM32),)
            if self.protocol_version == BootVersion.CF2_PROTO_VER:
                connected += (self.get_target(TargetTypes.NRF51),)
            info_cb(self.protocol_version, connected)

        if filename is not None:
            self.flash(filename, targets, cf, enable_console_log=enable_console_log, boot_delay=boot_delay)
            self.reset_to_firmware(boot_delay=5.0)

    def _get_flash_artifacts_from_zip(self, filename):
        if not zipfile.is_zipfile(filename):
            return []

        zf = zipfile.ZipFile(filename)

        manifest = zf.read('manifest.json').decode('utf8')
        manifest = json.loads(manifest)

        if manifest['version'] > 2:
            raise Exception('Wrong manifest version')

        print('Found manifest version: {}'.format(manifest['version']))

        flash_artifacts = []
        add_legacy_nRF51_s110 = False
        for (file, metadata) in manifest['files'].items():
            content = zf.read(file)

            # Handle version 1 of manifest where prerequisites for nRF soft-devices are not specified
            requires = [] if 'requires' not in metadata else metadata['requires']
            provides = [] if 'provides' not in metadata else metadata['provides']

            # Support both single target (string) and multiple targets (list)
            target_value = metadata['target']
            target_list = target_value if isinstance(target_value, list) else [target_value]

            for target_name in target_list:
                logger.debug('Processing target: {}'.format(target_name))
                # Create copies for each target to avoid shared mutable state
                target_requires = requires.copy()
                target_provides = provides.copy()

                # Check for legacy nRF51 requirement
                if len(target_requires) == 0 and target_name == 'nrf51' and metadata['type'] == 'fw':
                    target_requires.append('sd-s110')
                    # If there is no requires for the nRF51 fw target then we also need the legacy s110
                    # so add this to the file list afterwards
                    add_legacy_nRF51_s110 = True

                target = Target(metadata['platform'], target_name, metadata['type'],
                                target_provides, target_requires)
                flash_artifacts.append(FlashArtifact(content, target, metadata['release']))

        if add_legacy_nRF51_s110:
            print('Legacy format detected for manifest, adding s110+bl binary from distro')
            from importlib.resources import read_binary
            content = read_binary('cflib.resources.binaries', 'nrf51-s110-and-bl.bin')
            target = Target('cf2', 'nrf51', 'bootloader+softdevice', ['sd-s110'], [])
            release = None
            flash_artifacts.append(FlashArtifact(content, target, release))

        return flash_artifacts

    def _flash_flash(self, artifacts: List[FlashArtifact], targets: List[Target]):
        for (i, artifact) in enumerate(artifacts):
            self._internal_flash(artifact, i + 1, len(artifacts))

    def reset_to_firmware(self, boot_delay: float = 0.0) -> bool:
        status = False
        if self._cload.protocol_version == BootVersion.CF2_PROTO_VER:
            status = self._cload.reset_to_firmware(TargetTypes.NRF51, boot_delay)
        else:
            status = self._cload.reset_to_firmware(TargetTypes.STM32, boot_delay)

        if status:
            self.close()

        return status

    def close(self):
        if self._cload:
            self._cload.close()
            self._cload.link = None

    def _internal_flash(self, artifact: FlashArtifact, current_file_number=1, total_files=1, page_override=None):

        target_info = self._cload.targets[TargetTypes.from_string(artifact.target.target)]

        image = artifact.content
        t_data = target_info

        start_page = target_info.start_page
        if page_override is not None:
            start_page = page_override

        # If used from a UI we need some extra things for reporting progress
        factor = (100.0 * t_data.page_size) / len(image)
        progress = 0

        type_of_binary = 'Firmware'
        if artifact.target.type == 'bootloader+softdevice':
            type_of_binary = 'Bootloader+softdevice'

        if self.progress_cb:
            self.progress_cb(
                '{} ({}/{}) Starting...'.format(type_of_binary, current_file_number, total_files),
                int(progress))
        else:
            sys.stdout.write(
                'Flashing {} of {} to {} ({}): '.format(
                    current_file_number, total_files,
                    TargetTypes.to_string(t_data.id), artifact.target.type))
            sys.stdout.flush()

        if len(image) > ((t_data.flash_pages - start_page) *
                         t_data.page_size):
            if self.progress_cb:
                self.progress_cb('Error: Not enough space to flash the image file.', int(progress))
            else:
                print('Error: Not enough space to flash the image file.')
            raise Exception('Not enough space to flash the image file')

        if not self.progress_cb:
            logger.info(('%d bytes (%d pages) ' % (
                (len(image) - 1), int(len(image) / t_data.page_size) + 1)))
            sys.stdout.write(('%d bytes (%d pages) ' % (
                (len(image) - 1), int(len(image) / t_data.page_size) + 1)))
            sys.stdout.flush()

        # For each page
        ctr = 0  # Buffer counter
        for i in range(0, int((len(image) - 1) / t_data.page_size) + 1):
            if self.terminate_flashing_cb and self.terminate_flashing_cb():
                raise Exception('Flashing terminated')

            # Load the buffer
            if ((i + 1) * t_data.page_size) > len(image):
                self._cload.upload_buffer(
                    t_data.addr, ctr, 0, image[i * t_data.page_size:])
            else:
                self._cload.upload_buffer(
                    t_data.addr, ctr, 0,
                    image[i * t_data.page_size: (i + 1) * t_data.page_size])

            ctr += 1

            if self.progress_cb:
                progress += factor
                self.progress_cb('{} ({}/{}) Uploading buffer to {}...'.format(
                    type_of_binary,
                    current_file_number,
                    total_files,
                    TargetTypes.to_string(t_data.id)),

                    int(progress))
            else:
                sys.stdout.write('.')
                sys.stdout.flush()

            # Flash when the complete buffers are full
            if ctr >= t_data.buffer_pages:
                if self.progress_cb:
                    self.progress_cb('{} ({}/{}) Writing buffer to {}...'.format(
                        type_of_binary,
                        current_file_number,
                        total_files,
                        TargetTypes.to_string(t_data.id)),

                        int(progress))
                else:
                    sys.stdout.write('%d' % ctr)
                    sys.stdout.flush()
                if not self._cload.write_flash(t_data.addr, 0,
                                               start_page + i - (ctr - 1),
                                               ctr):
                    if self.progress_cb:
                        self.progress_cb(
                            'Error during flash operation (code {})'.format(
                                self._cload.error_code),
                            int(progress))
                    else:
                        print('\nError during flash operation (code %d). '
                              'Maybe wrong radio link?' %
                              self._cload.error_code)
                    raise Exception()

                ctr = 0

        if ctr > 0:
            if self.progress_cb:
                self.progress_cb('{} ({}/{}) Writing buffer to {}...'.format(
                    type_of_binary,
                    current_file_number,
                    total_files,
                    TargetTypes.to_string(t_data.id)),
                    int(progress))
            else:
                sys.stdout.write('%d' % ctr)
                sys.stdout.flush()
            if not self._cload.write_flash(
                    t_data.addr, 0,
                    (start_page + (int((len(image) - 1) / t_data.page_size)) -
                     (ctr - 1)), ctr):
                if self.progress_cb:
                    self.progress_cb(
                        'Error during flash operation (code {})'.format(
                            self._cload.error_code),
                        int(progress))
                else:
                    print('\nError during flash operation (code %d). Maybe'
                          ' wrong radio link?' % self._cload.error_code)
                raise Exception()

        sys.stdout.write('\n')
        sys.stdout.flush()

    def _get_platform_id(self):
        """Get platform identifier used in the zip manifest for curr copter"""
        identifier = 'cf1'
        if (BootVersion.is_cf2(self.protocol_version)):
            identifier = 'cf2'

        return identifier

    def console_callback(self, text: str):
        '''A callback to run when we get console text from Crazyflie'''
        # We do not add newlines to the text received, we get them from the
        # Crazyflie at appropriate places.
        print(text, end='')

    def _flash_deck_incrementally(self, artifacts: List[FlashArtifact], targets: List[Target], start_index: int,
                                  enable_console_log: Optional[bool] = False, boot_delay=0.0):
        flash_all_targets = len(targets) == 0
        if self.progress_cb:
            self.progress_cb('Identifying deck to be updated', 0)

        with SyncCrazyflie(self.clink, cf=Crazyflie()) as scf:
            if enable_console_log:
                scf.cf.console.receivedChar.add_callback(self.console_callback)

            deck_mems = scf.cf.mem.get_mems(MemoryElement.TYPE_DECK_MEMORY)
            deck_mems_count = len(deck_mems)
            if deck_mems_count == 0:
                return -1

            mgr = deck_memory.SyncDeckMemoryManager(deck_mems[0])
            try:
                decks = mgr.query_decks()
            except RuntimeError as e:
                if self.progress_cb:
                    message = f'Failed to read decks: {str(e)}'
                    self.progress_cb(message, 0)
                    logger.error(message)
                    time.sleep(2)
                    raise RuntimeError(message)

            # The decks dictionary uses the deck index as the key. Note that all indexes might not be present as some
            # decks do not support memory operations. decks.keys() might return the set {2, 4}.

            for deck_index in sorted(decks.keys()):
                deck = decks[deck_index]

                if self.terminate_flashing_cb and self.terminate_flashing_cb():
                    raise Exception('Flashing terminated')

                # Skip decks up to the start_index
                if deck_index < start_index:
                    continue

                # Check that we want to flash this deck
                deck_target = [t for t in targets if t == Target('deck', deck.name, 'fw', [], [])]
                if (not flash_all_targets) and len(deck_target) == 0:
                    print(f'Skipping {deck.name}, not in the target list')
                    continue

                # Check that we have an artifact for this deck
                deck_artifacts = [a for a in artifacts if a.target == Target('deck', deck.name, 'fw', [], [])]
                if len(deck_artifacts) == 0:
                    print(f'Skipping {deck.name}, no artifact for it in the .zip')
                    continue
                deck_artifact = deck_artifacts[0]

                if self.progress_cb:
                    self.progress_cb(f'Updating deck {deck.name}', 0)

                # Test and wait for the deck to be started
                timeout_time = time.time() + 4. + boot_delay
                while not deck.is_started:
                    if time.time() > timeout_time:
                        raise RuntimeError(f'Deck {deck.name} did not start')
                    print('Deck not yet started ...')
                    time.sleep(0.5)
                    deck = mgr.query_decks()[deck_index]

                # Supports upgrades?
                if not deck.supports_fw_upgrade:
                    print(f'Deck {deck.name} does not support firmware update, skipping!')
                    continue

                # Reset to bootloader mode, if supported
                if deck.supports_reset_to_bootloader:
                    print(f'Deck {deck.name}, reset to bootloader')
                    deck.reset_to_bootloader()

                    time.sleep(0.1)
                    deck = mgr.query_decks()[deck_index]
                else:
                    # Is an upgrade required?
                    if not deck.is_fw_upgrade_required:
                        print(f'Deck {deck.name} firmware up to date, skipping')
                        continue

                # Wait for bootloader to be ready
                timeout_time = time.time() + 4. + boot_delay
                while not deck.is_bootloader_active:
                    if time.time() > timeout_time:
                        raise RuntimeError(f'Deck {deck.name} did not enter bootloader mode')
                    print(f'Error: Deck {deck.name} bootloader not active yet...')
                    time.sleep(0.5)
                    deck = mgr.query_decks()[deck_index]

                progress_cb = self.progress_cb
                if not progress_cb:
                    def progress_cb(msg: str, percent: int):
                        frames = ['|', '/', '-', '\\']
                        frame = frames[int(percent) % 4]
                        print(f'{frame} {percent}% {msg}')

                # Flash the new firmware
                deck.set_fw_new_flash_size(len(deck_artifact.content))
                result = deck.write_sync(0, deck_artifact.content, progress_cb)
                if result:
                    if self.progress_cb:
                        self.progress_cb(f'Deck {deck.name} updated successful!', 100)
                else:
                    if self.progress_cb:
                        self.progress_cb(f'Failed to update deck {deck.name}', int(0))
                    raise RuntimeError(f'Failed to update deck {deck.name}')

                if enable_console_log:
                    # Wait a bit to let the console log print
                    time.sleep(6)

                # We flashed a deck, return for re-boot
                next_index = deck_index + 1
                return next_index

            # We have flashed the last deck
            return -1



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging

logger = logging.getLogger(__name__)


class MemoryElement(object):
    """A memory """

    TYPE_I2C = 0
    TYPE_1W = 1
    TYPE_DRIVER_LED = 0x10
    TYPE_LOCO = 0x11
    TYPE_TRAJ = 0x12
    TYPE_LOCO2 = 0x13
    TYPE_LH = 0x14
    TYPE_MEMORY_TESTER = 0x15
    TYPE_DRIVER_LEDTIMING = 0x17
    TYPE_APP = 0x18
    TYPE_DECK_MEMORY = 0x19
    TYPE_DECK_MULTIRANGER = 0x1A
    TYPE_DECK_PAA3905 = 0x1B

    def __init__(self, id, type, size, mem_handler):
        """Initialize the element with default values"""
        self.id = id
        self.type = type
        self.size = size
        self.mem_handler = mem_handler

    @staticmethod
    def type_to_string(t):
        """Get string representation of memory type"""
        if t == MemoryElement.TYPE_I2C:
            return 'I2C'
        if t == MemoryElement.TYPE_1W:
            return '1-wire'
        if t == MemoryElement.TYPE_DRIVER_LEDTIMING:
            return 'LED memory driver'
        if t == MemoryElement.TYPE_DRIVER_LED:
            return 'LED driver'
        if t == MemoryElement.TYPE_LOCO:
            return 'Loco Positioning'
        if t == MemoryElement.TYPE_TRAJ:
            return 'Trajectory'
        if t == MemoryElement.TYPE_LOCO2:
            return 'Loco Positioning 2'
        if t == MemoryElement.TYPE_LH:
            return 'Lighthouse positioning'
        if t == MemoryElement.TYPE_MEMORY_TESTER:
            return 'Memory tester'
        return 'Unknown'

    def new_data(self, mem, addr, data):
        logger.debug('New data, but not OW mem')

    def __str__(self):
        """Generate debug string for memory"""
        return ('Memory: id={}, type={}, size={}'.format(
            self.id, MemoryElement.type_to_string(self.type), self.size))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Bootloading utilities for the Crazyflie.
"""

__author__ = 'Bitcraze AB'
__all__ = ['BootVersion', 'TargetTypes', 'Target']


class BootVersion:
    CF1_PROTO_VER_0 = 0x00
    CF1_PROTO_VER_1 = 0x01
    CF2_PROTO_VER = 0x10

    @staticmethod
    def to_ver_string(ver):
        if (ver == BootVersion.CF1_PROTO_VER_0 or ver == BootVersion.
                CF1_PROTO_VER_1):
            return 'Crazyflie Nano Quadcopter (1.0)'
        if ver == BootVersion.CF2_PROTO_VER:
            return 'Crazyflie 2.0'
        return 'Unknown'

    @staticmethod
    def is_cf2(ver):
        return ver == BootVersion.CF2_PROTO_VER


class TargetTypes:
    STM32 = 0xFF
    NRF51 = 0xFE

    @staticmethod
    def to_string(target):
        if target == TargetTypes.STM32:
            return 'stm32'
        if target == TargetTypes.NRF51:
            return 'nrf51'
        return 'Unknown'

    @staticmethod
    def from_string(name):
        if name == 'stm32':
            return TargetTypes.STM32
        if name == 'nrf51':
            return TargetTypes.NRF51
        return 0


class Target:

    def __init__(self, id):
        self.id = id
        self.protocol_version = 0xFF
        self.page_size = 0
        self.buffer_pages = 0
        self.flash_pages = 0
        self.start_page = 0
        self.version = None
        self.cpuid = ''
        self.data = None

    def __str__(self):
        ret = ''
        ret += 'Target info: {} (0x{:X})\n'.format(
            TargetTypes.to_string(self.id), self.id)
        ret += 'Flash pages: {} | Page size: {} | Buffer pages: {} |' \
               ' Start page: {} | Version: {} \n'.format(self.flash_pages, self.page_size,
                                                         self.buffer_pages, self.start_page, self.version)
        ret += '%d KBytes of flash available for firmware image.' % (
            (self.flash_pages - self.start_page) * self.page_size / 1024)
        return ret



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2020 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
import struct

from .memory_element import MemoryElement

logger = logging.getLogger(__name__)


class MemoryTester(MemoryElement):
    """
    Memory interface for testing the memory sub system, end to end.

    Usage
    1. To verify reading:
      * Call read_data()
      * Wait for the callback to be called
      * Verify that readValidationSuccess is True

    2. To verify writing:
      * Set the parameter 'memTst.resetW' in the CF
      * call write_data()
      * Wait for the callback
      * Read the log var 'memTst.errCntW' from the CF and validate that it
        is 0
    """

    def __init__(self, id, type, size, mem_handler):
        """Initialize Memory tester"""
        super(MemoryTester, self).__init__(id=id, type=type, size=size,
                                           mem_handler=mem_handler)

        self._update_finished_cb = None
        self._write_finished_cb = None

        self.readValidationSuccess = True

    def new_data(self, mem, start_address, data):
        """Callback for when new memory data has been fetched"""
        if mem.id == self.id:
            for i in range(len(data)):
                actualValue = struct.unpack('<B', data[i:i + 1])[0]
                expectedValue = (start_address + i) & 0xff

                if (actualValue != expectedValue):
                    address = start_address + i
                    self.readValidationSuccess = False
                    logger.error(
                        'Error in data - expected: {}, actual: {}, address:{}',
                        expectedValue, actualValue, address)

                if self._update_finished_cb:
                    self._update_finished_cb(self)
                    self._update_finished_cb = None

    def read_data(self, start_address, size, update_finished_cb):
        """Request an update of the memory content"""
        if not self._update_finished_cb:
            self._update_finished_cb = update_finished_cb
            logger.debug('Reading memory {}'.format(self.id))
            self.mem_handler.read(self, start_address, size)

    def write_data(self, start_address, size, write_finished_cb):
        """Write data to the Crazyflie"""
        self._write_finished_cb = write_finished_cb
        data = bytearray()

        for i in range(size):
            value = (start_address + i) & 0xff
            data += struct.pack('<B', value)

        self.mem_handler.write(self, start_address, data, flush_queue=True)

    def write_done(self, mem, addr):
        if self._write_finished_cb and mem.id == self.id:
            logger.debug('Write of data finished')
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None

    def disconnect(self):
        self._update_finished_cb = None
        self._write_finished_cb = None



#!/usr/bin/env python3
# Demo that makes one Crazyflie take off 30cm above the first controller found
# Using the controller trigger it is then possible to 'grab' the Crazyflie
# and to make it move.
# If the Crazyflie has a ledring attached, the touchpad of the controller can
# be used to change the color of the led-ring LEDs
import colorsys
import math
import sys
import time

import openvr

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils.reset_estimator import reset_estimator

# URI to the Crazyflie to connect to
uri = 'radio://0/80/2M/E7E7E7E7E7'

print('Openning')
vr = openvr.init(openvr.VRApplication_Other)
print('Oppened')

# Find first controller or tracker
controllerId = None
poses = vr.getDeviceToAbsoluteTrackingPose(openvr.TrackingUniverseStanding, 0,
                                           openvr.k_unMaxTrackedDeviceCount)
for i in range(openvr.k_unMaxTrackedDeviceCount):
    if poses[i].bPoseIsValid:
        device_class = vr.getTrackedDeviceClass(i)
        if device_class == openvr.TrackedDeviceClass_Controller or\
           device_class == openvr.TrackedDeviceClass_GenericTracker:
            controllerId = i
            break

if controllerId is None:
    print('Cannot find controller or tracker, exiting')
    sys.exit(1)


def position_callback(timestamp, data, logconf):
    x = data['kalman.stateX']
    y = data['kalman.stateY']
    z = data['kalman.stateZ']
    print('pos: ({}, {}, {})'.format(x, y, z))


def start_position_printing(scf):
    log_conf = LogConfig(name='Position', period_in_ms=500)
    log_conf.add_variable('kalman.stateX', 'float')
    log_conf.add_variable('kalman.stateY', 'float')
    log_conf.add_variable('kalman.stateZ', 'float')

    scf.cf.log.add_config(log_conf)
    log_conf.data_received_cb.add_callback(position_callback)
    log_conf.start()


def vector_substract(v0, v1):
    return [v0[0] - v1[0], v0[1] - v1[1], v0[2] - v1[2]]


def vector_add(v0, v1):
    return [v0[0] + v1[0], v0[1] + v1[1], v0[2] + v1[2]]


def run_sequence(scf):
    cf = scf.cf

    cf.param.set_value('ring.effect', '7')

    poses = vr.getDeviceToAbsoluteTrackingPose(
        openvr.TrackingUniverseStanding, 0, openvr.k_unMaxTrackedDeviceCount)

    controller_pose = poses[controllerId]
    pose = controller_pose.mDeviceToAbsoluteTracking
    # setpoint = [-1*pose[2][3], -1*pose[0][3], pose[1][3] + 0.3]
    setpoint = [0, 0, 0.5]

    grabbed = False
    grab_controller_start = [0, 0, 0]
    grab_setpoint_start = [0, 0, 0]

    while True:
        poses = vr.getDeviceToAbsoluteTrackingPose(
            openvr.TrackingUniverseStanding, 0,
            openvr.k_unMaxTrackedDeviceCount)
        controller_state = vr.getControllerState(controllerId)[1]

        trigger = ((controller_state.ulButtonPressed & 0x200000000) != 0)
        trackpad_touched = ((controller_state.ulButtonTouched & 0x100000000) != 0)  # noqa

        trackpad = (0, 0)
        for i in range(openvr.k_unControllerStateAxisCount):
            axis_type = vr.getInt32TrackedDeviceProperty(
                controllerId, openvr.Prop_Axis0Type_Int32 + i)[0]
            if axis_type == openvr.k_eControllerAxis_TrackPad:
                trackpad = (controller_state.rAxis[i].x,
                            controller_state.rAxis[i].y)

        controller_pose = poses[controllerId]
        pose = controller_pose.mDeviceToAbsoluteTracking

        if trackpad_touched:
            angle = math.atan2(trackpad[1], trackpad[0])
            hue = (angle + math.pi) / (2*math.pi)
            rgb = colorsys.hsv_to_rgb(hue, 1, 0.3)

            print(rgb)

            cf.param.set_value('ring.solidRed',
                               '{}'.format(int(rgb[0] * 255)))
            cf.param.set_value('ring.solidGreen',
                               '{}'.format(int(rgb[1] * 255)))
            cf.param.set_value('ring.solidBlue',
                               '{}'.format(int(rgb[2] * 255)))

        if not grabbed and trigger:
            print('Grab started')
            grab_controller_start = [-1*pose[2][3], -1*pose[0][3], pose[1][3]]
            grab_setpoint_start = setpoint

        if grabbed and not trigger:
            print('Grab ended')

        grabbed = trigger

        if trigger:
            curr = [-1*pose[2][3], -1*pose[0][3], pose[1][3]]
            setpoint = vector_add(grab_setpoint_start,
                                  vector_substract(curr, grab_controller_start)
                                  )

        cf.commander.send_position_setpoint(setpoint[0],
                                            setpoint[1],
                                            setpoint[2],
                                            0)
        time.sleep(0.02)

    cf.commander.send_setpoint
    (0, 0, 0, 0)
    # Make sure that the last packet leaves before the link is closed
    # since the message queue is not flushed before closing
    time.sleep(0.1)


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
        reset_estimator(scf)
        # start_position_printing(scf)

        # Arm the Crazyflie
        scf.cf.platform.send_arming_request(True)
        time.sleep(1.0)

        run_sequence(scf)

    openvr.shutdown()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Crazyflie radio bootloader for flashing firmware.
"""
import binascii
import logging
import math
import struct
import time

import cflib.crtp
from .boottypes import Target
from .boottypes import TargetTypes
from cflib.crtp.crtpstack import CRTPPacket

__author__ = 'Bitcraze AB'
__all__ = ['Cloader']

logger = logging.getLogger(__name__)


class Cloader:
    """Bootloader utility for the Crazyflie"""

    def __init__(self, link, info_cb=None, in_boot_cb=None):
        """Init the communication class by starting to communicate with the
        link given. clink is the link address used after resetting to the
        bootloader.

        The device is actually considered in firmware mode.
        """
        self.link = None
        self.uri = link
        self.in_loader = False

        self.page_size = 0
        self.buffer_pages = 0
        self.flash_pages = 0
        self.start_page = 0
        self.cpuid = 'N/A'
        self.error_code = 0
        self.protocol_version = 0xFF

        self._info_cb = info_cb
        self._in_boot_cb = in_boot_cb

        self.targets = {}
        self.mapping = None
        self._available_boot_uri = ('radio://0/110/2M/E7E7E7E7E7', 'radio://0/0/2M/E7E7E7E7E7')

    def close(self):
        """ Close the link """
        if self.link:
            self.link.close()

    def scan_for_bootloader(self):
        link = cflib.crtp.get_link_driver('radio://0/80/2M/E7E7E7E7E7')
        ts = time.time()
        res = ()
        while len(res) == 0 and (time.time() - ts) < 10:
            res = link.scan_selected(self._available_boot_uri)

        link.close()

        if len(res) > 0:
            return res[0]
        return None

    def reset_to_bootloader(self, target_id: int) -> bool:
        pk = CRTPPacket(0xFF, [target_id, 0xFF])
        self.link.send_packet(pk)
        address = None

        timeout = 5  # seconds
        ts = time.time()
        while time.time() - ts < timeout:
            pk = self.link.receive_packet(2)
            if pk is None:
                continue
            if pk.port == 15 and pk.channel == 3 and len(pk.data) > 3:
                if struct.unpack('<BB', pk.data[0:2]) != (target_id, 0xFF):
                    continue

                address = 'B1' + binascii.hexlify(pk.data[2:6][::-1]).upper().decode('utf8')

                pk = CRTPPacket(0xFF, [target_id, 0xF0, 0x00])
                self.link.send_packet(pk)
                time.sleep(0.5)

                self.link.close()
                self.link = cflib.crtp.get_link_driver(f'radio://0/0/2M/{address}?safelink=0')
                time.sleep(0.5)
                return True
        self.link.close()
        return False

    def reset_to_firmware(self, target_id: int, boot_delay: float = 0.0) -> bool:
        """ Reset to firmware
        The parameter target_id corresponds to the device to reset.

        Return True if the reset has been done, False on timeout
        """
        pk = CRTPPacket(0xFF, [target_id, 0xFF])
        self.link.send_packet(pk)

        timeout = 5  # seconds
        ts = time.time()
        while time.time() - ts < timeout:
            answer = self.link.receive_packet(2)
            if answer is None:
                self.link.send_packet(pk)
                continue
            if answer.port == 15 and answer.channel == 3 and len(answer.data) > 2:
                if struct.unpack('<BB', pk.data[0:2]) != (target_id, 0xFF):
                    continue
                pk = CRTPPacket(0xff, [target_id, 0xf0, 0x01])
                self.link.send_packet(pk)
                time.sleep(1 + boot_delay)
                return True

        time.sleep(0.1)
        return False

    def open_bootloader_uri(self, uri=None):
        if self.link:
            self.link.close()
        if uri:
            self.link = cflib.crtp.get_link_driver(uri + '?safelink=0')
        else:
            self.link = cflib.crtp.get_link_driver(
                self.clink_address + '?safelink=0')

    def check_link_and_get_info(self, target_id=0xFF):
        """Try to get a connection with the bootloader ...
           update_info has a timeout of 10 seconds """
        if self._update_info(target_id):
            if self._in_boot_cb:
                self._in_boot_cb.call(True, self.targets[
                    target_id].protocol_version)
            if self._info_cb:
                self._info_cb.call(self.targets[target_id])
            return True
        return False

    def request_info_update(self, target_id):
        if target_id not in self.targets:
            self._update_info(target_id)
        if self._info_cb:
            self._info_cb.call(self.targets[target_id])
        return self.targets[target_id]

    def _update_info(self, target_id):
        """ Call the command getInfo and fill up the information received in
        the fields of the object
        """

        # Call getInfo ...
        pk = CRTPPacket()
        pk.set_header(0xFF, 0xFF)
        pk.data = (target_id, 0x10)
        self.link.send_packet(pk)

        timeout = 10  # seconds
        ts = time.time()
        while time.time() - ts < timeout:
            # Wait for the answer
            answer = self.link.receive_packet(2)
            if answer is None:
                self.link.send_packet(pk)

            if (answer and answer.header == 0xFF and struct.unpack('<BB', answer.data[0:2]) ==
                    (target_id, 0x10)):
                tab = struct.unpack('BBHHHH', answer.data[0:10])
                cpuid = struct.unpack('B' * 12, answer.data[10:22])
                if target_id not in self.targets:
                    self.targets[target_id] = Target(target_id)
                self.targets[target_id].addr = target_id
                if len(answer.data) > 22:
                    self.targets[target_id].protocol_version = answer.data[22]
                    self.protocol_version = answer.data[22]
                if len(answer.data) > 23 and len(answer.data) > 26:
                    code_state = ''
                    if answer.data[24] & 0x80 != 0:
                        code_state = '+'
                    answer.data[24] &= 0x7F
                    major = struct.unpack('H', answer.data[23:25])[0]
                    minor = answer.data[25]
                    patch = answer.data[26]
                    self.targets[target_id].version = '{}.{}.{}{}'.format(major, minor, patch, code_state)
                self.targets[target_id].page_size = tab[2]
                self.targets[target_id].buffer_pages = tab[3]
                self.targets[target_id].flash_pages = tab[4]
                self.targets[target_id].start_page = tab[5]
                self.targets[target_id].cpuid = '%02X' % cpuid[0]
                for i in cpuid[1:]:
                    self.targets[target_id].cpuid += ':%02X' % i

                if (self.protocol_version == 0x10 and
                        target_id == TargetTypes.STM32):
                    self._update_mapping(target_id)

                return True

        return False

    def _update_mapping(self, target_id):
        pk = CRTPPacket()
        pk.set_header(0xff, 0xff)
        pk.data = (target_id, 0x12)
        self.link.send_packet(pk)

        pk = self.link.receive_packet(2)

        if (pk and pk.header == 0xFF and len(pk.data) >= 2 and struct.unpack('<BB', pk.data[0:2]) ==
                (target_id, 0x12)):
            m = pk.datat[2:]

            if (len(m) % 2) != 0:
                raise Exception('Malformed flash mapping packet')

            self.mapping = []
            page = 0
            for i in range(int(len(m) / 2)):
                for j in range(m[2 * i]):
                    self.mapping.append(page)
                    page += m[(2 * i) + 1]

    def upload_buffer(self, target_id, page, address, buff):
        """Upload data into a buffer on the Crazyflie"""
        # print len(buff)
        count = 0
        pk = CRTPPacket()
        pk.set_header(0xFF, 0xFF)
        pk.data = struct.pack('=BBHH', target_id, 0x14, page, address)

        for i in range(0, len(buff)):
            pk.data.append(buff[i])

            count += 1

            if count > 24:
                self.link.send_packet(pk)
                count = 0
                pk = CRTPPacket()
                pk.set_header(0xFF, 0xFF)
                pk.data = struct.pack('=BBHH', target_id, 0x14, page,
                                      i + address + 1)

        self.link.send_packet(pk)

    def read_flash(self, addr=0xFF, page=0x00):
        """Read back a flash page from the Crazyflie and return it"""
        buff = bytearray()

        page_size = self.targets[addr].page_size

        for i in range(0, int(math.ceil(page_size / 25.0))):
            pk = None
            retry_counter = 5
            while ((not pk or pk.header != 0xFF or
                    struct.unpack('<BBHH', pk.data[0:6]) != (addr, 0x1C, page, (i * 25))) and
                    retry_counter >= 0):
                pk = CRTPPacket()
                pk.set_header(0xFF, 0xFF)
                pk.data = struct.pack('<BBHH', addr, 0x1C, page, (i * 25))
                self.link.send_packet(pk)

                pk = self.link.receive_packet(1)
                retry_counter -= 1
            if (retry_counter < 0):
                return None
            else:
                buff += pk.data[6:]

        # For some reason we get one byte extra here...
        return buff[0:page_size]

    def write_flash(self, addr, page_buffer, target_page, page_count):
        """Initiate flashing of data in the buffer to flash."""
        # print "Write page", flashPage
        # print "Writing page [%d] and [%d] forward" % (flashPage, nPage)
        pk = None

        # Flushing downlink ...
        pk = self.link.receive_packet(0)
        while pk is not None:
            pk = self.link.receive_packet(0)

        retry_counter = 5
        # print "Flasing to 0x{:X}".format(addr)
        while ((not pk or pk.header != 0xFF or len(pk.data) < 2 or
                struct.unpack('<BB', pk.data[0:2]) != (addr, 0x18)) and
               retry_counter >= 0):
            pk = CRTPPacket()
            pk.set_header(0xFF, 0xFF)
            pk.data = struct.pack('<BBHHH', addr, 0x18, page_buffer,
                                  target_page, page_count)
            self.link.send_packet(pk)

            # Timeout for writing to flash is raised from 1s (used elsewhere
            # in this module) to 2.5s because it may take more than a second
            # to erase a page on the STM32F405.
            #
            # See https://github.com/bitcraze/crazyflie-lib-python/issues/98
            # for more details.
            pk = self.link.receive_packet(2.5)
            retry_counter -= 1

        if retry_counter < 0:
            self.error_code = -1
            return False

        self.error_code = pk.data[3]

        return pk.data[2] == 1

    def decode_cpu_id(self, cpuid):
        """Decode the CPU id into a string"""
        ret = ()
        for i in cpuid.split(':'):
            ret += (eval('0x' + i),)

        return ret



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2025 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Example of how to connect to a motion capture system and feed positions (only)
to multiple Crazyflies, using the motioncapture library.

The script uses the position high level and the motion commander to fly circles and waypoints.

Set the uri to the radio settings of your Crazyflies, set the rigid body names and
modify the mocap setting matching your system.
"""
import time
from threading import Thread

import motioncapture

import cflib.crtp
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.positioning.motion_commander import MotionCommander
from cflib.positioning.position_hl_commander import PositionHlCommander

# The type of the mocap system
# Valid options are: 'vicon', 'optitrack', 'optitrack_closed_source', 'qualisys', 'nokov', 'vrpn', 'motionanalysis'
mocap_system_type = 'optitrack'

# The host name or ip address of the mocap system
host_name = '10.223.0.31'

# Maps the URIs to the rigid-body names as streamed by, e.g., OptiTrack's Motive
swarm_config = [
    ('radio://0/80/2M/E7E7E7E7E7', 'cf1'),
    #    ('radio://0/80/2M/E7E7E7E7E8', 'cf2'),
    #    ('radio://0/80/2M/E7E7E7E7E9', 'cf3'),
    #   Add more URIs if you want more copters in the swarm
]

uris = [uri for uri, _ in swarm_config]
rbs = {uri: name for uri, name in swarm_config}


class MocapWrapper(Thread):
    def __init__(self, active_rbs_cfs):
        Thread.__init__(self)
        self.active_rbs_cfs = active_rbs_cfs
        self._stay_open = True
        self.counter = 0
        self.start()

    def close(self):
        self._stay_open = False

    def run(self):
        mc = motioncapture.connect(mocap_system_type, {'hostname': host_name})
        while self._stay_open:
            mc.waitForNextFrame()
            self.counter += 1
            for name, obj in mc.rigidBodies.items():
                if name in self.active_rbs_cfs:
                    pos = obj.position
                    # Only send positions
                    self.active_rbs_cfs[name].extpos.send_extpos(pos[0], pos[1], pos[2])
                    if self.counter == 200:
                        print(f'Sent pos {pos} for {name}')
            if self.counter == 200:
                self.counter = 0


def activate_kalman_estimator(scf: SyncCrazyflie):
    scf.cf.param.set_value('stabilizer.estimator', '2')

    # Set the std deviation for the quaternion data pushed into the
    # kalman filter. The default value seems to be a bit too low.
    scf.cf.param.set_value('locSrv.extQuatStdDev', 0.06)


def run_sequence(scf: SyncCrazyflie):
    print('This is: ', scf._link_uri)
    scf.cf.platform.send_arming_request(True)
    time.sleep(1.0)

    # .takeoff() is automatic when entering the "with PositionHlCommander" context
    if rbs[scf._link_uri] == 'cf1':
        with PositionHlCommander(scf, controller=PositionHlCommander.CONTROLLER_PID) as pc:
            pc.set_default_velocity(0.5)
            for i in range(3):  # fly a triangle with changing altitude
                pc.go_to(1.0, 1.0, 1.5)
                pc.go_to(1.0, -1.0, 1.5)
                pc.go_to(0.5, 0.0, 2.0)
            pc.go_to(0.5, 0.0, 0.15)
    elif rbs[scf._link_uri] == 'cf2':
        with PositionHlCommander(scf, controller=PositionHlCommander.CONTROLLER_PID) as pc:
            pc.set_default_velocity(0.3)
            for i in range(3):  # fly side to side
                pc.go_to(0.2, 1.0, 0.85)
                pc.go_to(0.2, -1.0, 0.85)
            pc.go_to(0.0, 0.0, 0.15)
    elif rbs[scf._link_uri] == 'cf3':
        with MotionCommander(scf) as mc:
            # 2 right loops and 2 left loops
            mc.back(0.8)
            time.sleep(1)
            mc.up(0.5)
            time.sleep(1)
            mc.circle_right(0.5, velocity=0.4, angle_degrees=720)
            time.sleep(1)
            mc.up(0.5)
            time.sleep(1)
            mc.circle_left(0.5, velocity=0.4, angle_degrees=720)
            time.sleep(1)
            mc.down(0.5)
    # .land() is automatic when exiting the scope of context "with PositionHlCommander"


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    factory = CachedCfFactory(rw_cache='./cache')
    with Swarm(uris, factory=factory) as swarm:
        active_rbs_cfs = {rbs[uri]: scf.cf for uri, scf in swarm._cfs.items()}
        mocap_thread = MocapWrapper(active_rbs_cfs)

        swarm.parallel_safe(activate_kalman_estimator)
        time.sleep(1)
        swarm.reset_estimators()
        time.sleep(2)
        swarm.parallel_safe(run_sequence)

    mocap_thread.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
# flake8: noqa
"""
Enables logging of variables from the Crazyflie.

When a Crazyflie is connected it's possible to download a TableOfContent of all
the variables that can be logged. Using this it's possible to add logging
configurations where selected variables are sent to the client at a
specified period.

#### Terminology

| Term              | Description |
| ----------------- | ----------- |
| <nobr>Log configuration</nobr> | A configuration with a period and a number of variables that are present in the TOC. |
| Stored as         | The size and type of the variable as declared in the Crazyflie firmware. |
| Fetch as          | The size and type that a variable should be fetched as. This does not have to be the same as the size and type it's stored as. |

#### States of a configuration

| State             | Description |
| ----------------- | ----------- |
| <nobr>Created on host</nobr> | When a configuration is created the contents is checked so that all the variables are present in the TOC. If not then the configuration cannot be created. |
| <nobr>Created on CF</nobr>   | When the configuration is deemed valid it is added to the Crazyflie. At this time the memory constraint is checked and the status returned. |
| <nobr>Started on CF</nobr>   | Any added block that is not started can be started. Once started the Crazyflie will send back logdata periodically according to the specified period when it's created. |
| <nobr>Stopped on CF</nobr>   | Any started configuration can be stopped. The memory taken by the configuration on the Crazyflie is NOT freed, the only effect is that the Crazyflie will stop sending logdata back to the host. |
| <nobr>Deleted on CF</nobr>   | Any block that is added can be deleted. When this is done the memory taken by the configuration is freed on the Crazyflie. The configuration will have to be re-added to be used again. |
"""
import errno
import logging
import struct

from .toc import Toc
from .toc import TocFetcher
from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort
from cflib.utils.callbacks import Caller

__author__ = 'Bitcraze AB'
__all__ = ['Log', 'LogTocElement']

# Channels used for the logging port
CHAN_TOC = 0
CHAN_SETTINGS = 1
CHAN_LOGDATA = 2

# Commands used when accessing the Table of Contents
CMD_TOC_ELEMENT = 0  # original version: up to 255 entries
CMD_TOC_INFO = 1    # original version: up to 255 entries
CMD_GET_ITEM_V2 = 2  # version 2: up to 16k entries
CMD_GET_INFO_V2 = 3  # version 2: up to 16k entries

# Commands used when accessing the Log configurations
CMD_CREATE_BLOCK = 0
CMD_APPEND_BLOCK = 1
CMD_DELETE_BLOCK = 2
CMD_START_LOGGING = 3
CMD_STOP_LOGGING = 4
CMD_RESET_LOGGING = 5
CMD_CREATE_BLOCK_V2 = 6
CMD_APPEND_BLOCK_V2 = 7

# Possible states when receiving TOC
IDLE = 'IDLE'
GET_TOC_INF = 'GET_TOC_INFO'
GET_TOC_ELEMENT = 'GET_TOC_ELEMENT'


logger = logging.getLogger(__name__)


class LogVariable():
    """A logging variable"""

    TOC_TYPE = 0
    MEM_TYPE = 1

    def __init__(self, name='', fetchAs='uint8_t', varType=TOC_TYPE,
                 storedAs='', address=0):
        self.name = name
        self.fetch_as = LogTocElement.get_id_from_cstring(fetchAs)
        if (len(storedAs) == 0):
            self.stored_as = self.fetch_as
        else:
            self.stored_as = LogTocElement.get_id_from_cstring(storedAs)
        self.address = address
        self.type = varType
        self.stored_as_string = storedAs
        self.fetch_as_string = fetchAs

    def is_toc_variable(self):
        """
        Return true if the variable should be in the TOC, false if raw memory
        variable
        """
        return self.type == LogVariable.TOC_TYPE

    def get_storage_and_fetch_byte(self):
        """Return what the variable is stored as and fetched as"""
        return (self.fetch_as | (self.stored_as << 4))

    def __str__(self):
        return ('LogVariable: name=%s, store=%s, fetch=%s' %
                (self.name, LogTocElement.get_cstring_from_id(self.stored_as),
                 LogTocElement.get_cstring_from_id(self.fetch_as)))


class LogConfig(object):
    """Representation of one log configuration that enables logging
    from the Crazyflie"""

    # Maximum log payload length (4 bytes are used for block id and timestamp)
    MAX_LEN = 26

    def __init__(self, name, period_in_ms):
        """Initialize the entry"""
        self.data_received_cb = Caller()
        self.error_cb = Caller()
        self.started_cb = Caller()
        self.added_cb = Caller()
        self.err_no = 0

        # These 3 variables are set by the log subsystem when the bock is added
        self.id = 0
        self.cf = None
        self.useV2 = False

        self.period = int(period_in_ms / 10)
        self.period_in_ms = period_in_ms
        self._added = False
        self._started = False
        self.pending = False
        self.valid = False
        self.variables = []
        self.default_fetch_as = []
        self.name = name

    def add_variable(self, name, fetch_as=None):
        """Add a new variable to the configuration.

        name - Complete name of the variable in the form group.name
        fetch_as - String representation of the type the variable should be
                   fetched as (i.e uint8_t, float, FP16, etc)

        If no fetch_as type is supplied, then the stored as type will be used
        (i.e the type of the fetched variable is the same as it's stored in the
        Crazyflie)."""
        if fetch_as:
            self.variables.append(LogVariable(name, fetch_as))
        else:
            # We cannot determine the default type until we have connected. So
            # save the name and we will add these once we are connected.
            self.default_fetch_as.append(name)

    def add_memory(self, name, fetch_as, stored_as, address):
        """Add a raw memory position to log.

        name - Arbitrary name of the variable
        fetch_as - String representation of the type of the data the memory
                   should be fetch as (i.e uint8_t, float, FP16)
        stored_as - String representation of the type the data is stored as
                    in the Crazyflie
        address - The address of the data
        """
        self.variables.append(LogVariable(name, fetch_as, LogVariable.MEM_TYPE,
                                          stored_as, address))

    def _set_added(self, added):
        if added != self._added:
            self.added_cb.call(self, added)
        self._added = added

    def _get_added(self):
        return self._added

    def _set_started(self, started):
        if started != self._started:
            self.started_cb.call(self, started)
        self._started = started

    def _get_started(self):
        return self._started

    added = property(_get_added, _set_added)
    started = property(_get_started, _set_started)

    def _cmd_create_block(self):
        if self.useV2:
            return CMD_CREATE_BLOCK_V2
        else:
            return CMD_CREATE_BLOCK

    def _cmd_append_block(self):
        if self.useV2:
            return CMD_APPEND_BLOCK_V2
        else:
            return CMD_APPEND_BLOCK

    def _setup_log_elements(self, pk, next_to_add):
        i = next_to_add
        for i in range(next_to_add, len(self.variables)):
            var = self.variables[i]
            if (var.is_toc_variable() is False):  # Memory location
                logger.debug('Logging to raw memory %d, 0x%04X',
                             var.get_storage_and_fetch_byte(), var.address)
                pk.data.append(struct.pack('<B',
                                           var.get_storage_and_fetch_byte()))
                pk.data.append(struct.pack('<I', var.address))
            else:  # Item in TOC
                element_id = self.cf.log.toc.get_element_id(var.name)
                logger.debug('Adding %s with id=%d and type=0x%02X',
                             var.name,
                             element_id,
                             var.get_storage_and_fetch_byte())
                pk.data.append(var.get_storage_and_fetch_byte())
                if self.useV2:
                    size_to_add = 2
                    if pk.available_data_size() >= size_to_add:
                        pk.data.append(element_id & 0x0ff)
                        pk.data.append((element_id >> 8) & 0x0ff)
                    else:
                        # Packet is full
                        return False, i
                else:
                    pk.data.append(element_id)

        return True, i

    def create(self):
        """Save the log configuration in the Crazyflie"""
        command = self._cmd_create_block()
        next_to_add = 0
        is_done = False

        num_variables = 0
        pending = 0
        for block in self.cf.log.log_blocks:
            if block.pending or block.added or block.started:
                pending += 1
                num_variables += len(block.variables)

        if pending < Log.MAX_BLOCKS:
            #
            # The Crazyflie firmware can only handle 128 variables before
            # erroring out with ENOMEM.
            #
            if num_variables + len(self.variables) > Log.MAX_VARIABLES:
                raise AttributeError(
                    ('Adding this configuration would exceed max number '
                     'of variables (%d)' % Log.MAX_VARIABLES)
                )
        else:
            raise AttributeError(
                'Configuration has max number of blocks (%d)' % Log.MAX_BLOCKS
            )
        self.pending += 1
        while not is_done:
            pk = CRTPPacket()
            pk.set_header(5, CHAN_SETTINGS)
            pk.data = (command, self.id)
            is_done, next_to_add = self._setup_log_elements(pk, next_to_add)

            logger.debug('Adding/appending log block id {}'.format(self.id))
            self.cf.send_packet(pk, expected_reply=(command, self.id))

            # Use append if we have to add more variables
            command = self._cmd_append_block()

    def start(self):
        """Start the logging for this entry"""
        if (self.cf.link is not None):
            if (self._added is False):
                self.create()
                logger.debug('First time block is started, add block')
            else:
                logger.debug('Block already registered, starting logging'
                             ' for id=%d', self.id)
                pk = CRTPPacket()
                pk.set_header(5, CHAN_SETTINGS)
                pk.data = (CMD_START_LOGGING, self.id, self.period)
                self.cf.send_packet(pk, expected_reply=(
                    CMD_START_LOGGING, self.id))

    def stop(self):
        """Stop the logging for this entry"""
        if (self.cf.link is not None):
            if (self.id is None):
                logger.warning('Stopping block, but no block registered')
            else:
                logger.debug('Sending stop logging for block id=%d', self.id)
                pk = CRTPPacket()
                pk.set_header(5, CHAN_SETTINGS)
                pk.data = (CMD_STOP_LOGGING, self.id)
                self.cf.send_packet(
                    pk, expected_reply=(CMD_STOP_LOGGING, self.id))

    def delete(self):
        """Delete this entry in the Crazyflie"""
        if (self.cf.link is not None):
            if (self.id is None):
                logger.warning('Delete block, but no block registered')
            else:
                logger.debug('LogEntry: Sending delete logging for block id=%d'
                             % self.id)
                pk = CRTPPacket()
                pk.set_header(5, CHAN_SETTINGS)
                pk.data = (CMD_DELETE_BLOCK, self.id)
                self.cf.send_packet(
                    pk, expected_reply=(CMD_DELETE_BLOCK, self.id))

    def unpack_log_data(self, log_data, timestamp):
        """Unpack received logging data so it represent real values according
        to the configuration in the entry"""
        ret_data = {}
        data_index = 0
        for var in self.variables:
            size = LogTocElement.get_size_from_id(var.fetch_as)
            name = var.name
            unpackstring = LogTocElement.get_unpack_string_from_id(
                var.fetch_as)
            value = struct.unpack(
                unpackstring, log_data[data_index:data_index + size])[0]
            data_index += size
            ret_data[name] = value
        self.data_received_cb.call(timestamp, ret_data, self)


class LogTocElement:
    """An element in the Log TOC."""
    types = {0x01: ('uint8_t', '<B', 1),
             0x02: ('uint16_t', '<H', 2),
             0x03: ('uint32_t', '<L', 4),
             0x04: ('int8_t', '<b', 1),
             0x05: ('int16_t', '<h', 2),
             0x06: ('int32_t', '<i', 4),
             0x08: ('FP16', '<e', 2),
             0x07: ('float', '<f', 4)}

    @staticmethod
    def get_id_from_cstring(name):
        """Return variable type id given the C-storage name"""
        for key in list(LogTocElement.types.keys()):
            if (LogTocElement.types[key][0] == name):
                return key
        raise KeyError('Type [%s] not found in LogTocElement.types!' % name)

    @staticmethod
    def get_cstring_from_id(ident):
        """Return the C-storage name given the variable type id"""
        try:
            return LogTocElement.types[ident][0]
        except KeyError:
            raise KeyError('Type [%d] not found in LogTocElement.types'
                           '!' % ident)

    @staticmethod
    def get_size_from_id(ident):
        """Return the size in bytes given the variable type id"""
        try:
            return LogTocElement.types[ident][2]
        except KeyError:
            raise KeyError('Type [%d] not found in LogTocElement.types'
                           '!' % ident)

    @staticmethod
    def get_unpack_string_from_id(ident):
        """Return the Python unpack string given the variable type id"""
        try:
            return LogTocElement.types[ident][1]
        except KeyError:
            raise KeyError(
                'Type [%d] not found in LogTocElement.types!' % ident)

    def __init__(self, ident=0, data=None):
        """TocElement creator. Data is the binary payload of the element."""

        self.ident = ident

        if (data):
            naming = data[1:]
            zt = bytearray((0, ))
            self.group = naming[:naming.find(zt)].decode('ISO-8859-1')
            self.name = naming[naming.find(zt) + 1:-1].decode('ISO-8859-1')

            self.ctype = LogTocElement.get_cstring_from_id(data[0])
            self.pytype = LogTocElement.get_unpack_string_from_id(data[0])

            self.access = data[0] & 0x10


class Log():
    """Create log configuration"""

    MAX_BLOCKS = 16
    MAX_VARIABLES = 128

    # These codes can be decoded using os.stderror, but
    # some of the text messages will look very strange
    # in the UI, so they are redefined here
    _err_codes = {
        errno.ENOMEM: 'No more memory available',
        errno.ENOEXEC: 'Command not found',
        errno.ENOENT: 'No such block id',
        errno.E2BIG: 'Block too large',
        errno.EEXIST: 'Block already exists'
    }

    def __init__(self, crazyflie=None):
        self.log_blocks = []
        # Called with newly created blocks
        self.block_added_cb = Caller()

        self.cf = crazyflie
        self.toc = None
        self.cf.add_port_callback(CRTPPort.LOGGING, self._new_packet_cb)

        self.toc_updated = Caller()
        self.state = IDLE
        self.fake_toc_crc = 0xDEADBEEF

        self._refresh_callback = None
        self._toc_cache = None

        self._config_id_counter = 1

        self._useV2 = False

    def add_config(self, logconf):
        """Add a log configuration to the logging framework.

        When doing this the contents of the log configuration will be validated
        and listeners for new log configurations will be notified. When
        validating the configuration the variables are checked against the TOC
        to see that they actually exist. If they don't then the configuration
        cannot be used. Since a valid TOC is required, a Crazyflie has to be
        connected when calling this method, otherwise it will fail."""

        if not self.cf.link:
            logger.error('Cannot add configs without being connected to a '
                         'Crazyflie!')
            return

        # If the log configuration contains variables that we added without
        # type (i.e we want the stored as type for fetching as well) then
        # resolve this now and add them to the block again.
        for name in logconf.default_fetch_as:
            var = self.toc.get_element_by_complete_name(name)
            if not var:
                logger.warning(
                    '%s not in TOC, this block cannot be used!', name)
                logconf.valid = False
                raise KeyError('Variable {} not in TOC'.format(name))
            # Now that we know what type this variable has, add it to the log
            # config again with the correct type
            logconf.add_variable(name, var.ctype)

        # Now check that all the added variables are in the TOC and that
        # the total size constraint of a data packet with logging data is
        # not
        size = 0
        for var in logconf.variables:
            size += LogTocElement.get_size_from_id(var.fetch_as)
            # Check that we are able to find the variable in the TOC so
            # we can return error already now and not when the config is sent
            if var.is_toc_variable():
                if (self.toc.get_element_by_complete_name(var.name) is None):
                    logger.warning(
                        'Log: %s not in TOC, this block cannot be used!',
                        var.name)
                    logconf.valid = False
                    raise KeyError('Variable {} not in TOC'.format(var.name))

        if (size <= LogConfig.MAX_LEN and
                (logconf.period > 0 and logconf.period < 0xFF)):
            logconf.valid = True
            logconf.cf = self.cf
            logconf.id = self._config_id_counter
            logconf.useV2 = self._useV2
            self._config_id_counter = (self._config_id_counter + 1) % 255
            self.log_blocks.append(logconf)
            self.block_added_cb.call(logconf)
        else:
            logconf.valid = False
            raise AttributeError(
                'The log configuration is too large or has an invalid '
                'parameter')

    def reset(self):
        """
        Reset the log system and remove all log blocks
        """
        self.log_blocks = []
        self._send_reset_packet()

    def refresh_toc(self, refresh_done_callback, toc_cache):
        """Start refreshing the table of loggale variables"""

        self._useV2 = self.cf.platform.get_protocol_version() >= 4

        self._toc_cache = toc_cache
        self._refresh_callback = refresh_done_callback
        self.toc = None

        self._send_reset_packet()

    def _send_reset_packet(self):
        pk = CRTPPacket()
        pk.set_header(CRTPPort.LOGGING, CHAN_SETTINGS)
        pk.data = (CMD_RESET_LOGGING,)
        self.cf.send_packet(pk, expected_reply=(CMD_RESET_LOGGING,))

    def _find_block(self, id):
        for block in self.log_blocks:
            if block.id == id:
                return block
        return None

    def _new_packet_cb(self, packet):
        """Callback for newly arrived packets with TOC information"""
        chan = packet.channel
        cmd = packet.data[0]
        payload = packet.data[1:]

        if (chan == CHAN_SETTINGS):
            id = payload[0]
            error_status = payload[1]
            block = self._find_block(id)
            if cmd == CMD_CREATE_BLOCK or cmd == CMD_CREATE_BLOCK_V2:
                if (block is not None):
                    if error_status == 0 or error_status == errno.EEXIST:
                        if not block.added:
                            logger.debug('Have successfully added id=%d', id)

                            pk = CRTPPacket()
                            pk.set_header(5, CHAN_SETTINGS)
                            pk.data = (CMD_START_LOGGING, id, block.period)
                            self.cf.send_packet(pk, expected_reply=(
                                CMD_START_LOGGING, id))
                            block.added = True
                            block.pending = False
                    else:
                        msg = self._err_codes[error_status]
                        logger.warning('Error %d when adding id=%d (%s)',
                                       error_status, id, msg)
                        block.err_no = error_status
                        block.added_cb.call(False)
                        block.error_cb.call(block, msg)

                else:
                    logger.warning('No LogEntry to assign block to !!!')
            if (cmd == CMD_START_LOGGING):
                if (error_status == 0x00):
                    logger.info('Have successfully started logging for id=%d',
                                id)
                    if block:
                        block.started = True

                else:
                    msg = self._err_codes[error_status]
                    logger.warning('Error %d when starting id=%d (%s)',
                                   error_status, id, msg)
                    if block:
                        block.err_no = error_status
                        block.started_cb.call(self, False)
                        # This is a temporary fix, we are adding a new issue
                        # for this. For some reason we get an error back after
                        # the block has been started and added. This will show
                        # an error in the UI, but everything is still working.
                        # block.error_cb.call(block, msg)

            if (cmd == CMD_STOP_LOGGING):
                if (error_status == 0x00):
                    logger.info('Have successfully stopped logging for id=%d',
                                id)
                    if block:
                        block.started = False

            if (cmd == CMD_DELETE_BLOCK):
                # Accept deletion of a block that isn't added. This could
                # happen due to timing (i.e add/start/delete in fast sequence)
                if error_status == 0x00 or error_status == errno.ENOENT:
                    logger.info('Have successfully deleted id=%d', id)
                    if block:
                        block.started = False
                        block.added = False

            if (cmd == CMD_RESET_LOGGING):
                # Guard against multiple responses due to re-sending
                if not self.toc:
                    logger.debug('Logging reset, continue with TOC download')
                    self.log_blocks = []

                    self.toc = Toc()
                    toc_fetcher = TocFetcher(self.cf, LogTocElement,
                                             CRTPPort.LOGGING,
                                             self.toc, self._refresh_callback,
                                             self._toc_cache)
                    toc_fetcher.start()

        if (chan == CHAN_LOGDATA):
            chan = packet.channel
            id = packet.data[0]
            block = self._find_block(id)
            timestamps = struct.unpack('<BBB', packet.data[1:4])
            timestamp = (
                timestamps[0] | timestamps[1] << 8 | timestamps[2] << 16)
            logdata = packet.data[4:]
            if (block is not None):
                block.unpack_log_data(logdata, timestamp)
            else:
                logger.warning('Error no LogEntry to handle id=%d', id)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from collections import namedtuple

import numpy as np
import numpy.typing as npt

from ._ippe import mat_run


class IppeCf:
    """
    A wrapper class to simplify usage of IPPE in CF code.
    Converts between CF style of coordinate systems/data structures and
    IPPE (Open CV) style.
    """

    # Rotation matrix to transform from IPPE to CF
    _R_ippe_to_cf = np.array([
        [0.0, 0.0, 1.0],
        [-1.0, 0.0, 0.0],
        [0.0, -1.0, 0.0],
    ])

    # Rotation matrix to transform from CF to IPPE
    _R_cf_to_ippe = np.transpose(_R_ippe_to_cf)

    Solution = namedtuple('Solution', ['R', 't', 'reproj_err'])

    @staticmethod
    def solve(U_cf: npt.ArrayLike, Q_cf: npt.ArrayLike) -> list[Solution]:
        """
        The solution to Perspective IPPE with point correspondences computed
        between points in world coordinates on the plane z=0, and normalised points in the
        camera's image.

        This is a wrapper function to convert from/to CF coordinate system/array style

        :param U_cf: Nx3 matrix holding the model points in world coordinates.
        :param Q_cf: Nx2 matrix holding the points in the image. These are in normalised
                     pixel coordinates. That is, the effects of the camera's intrinsic matrix
                     and lens distortion are corrected, so that the Q projects with a perfect
                     pinhole model.

                     First param: Y (positive to the left)
                     Second param: Z (positive up)
        :return: A list that contains 2 sets of pose solution from IPPE including rotation matrix
                 translation matrix, and reprojection error. The first solution in the list has
                 the smallest reprojection error.
        """

        U, Q = IppeCf._cf_to_ippe(U_cf, Q_cf)
        solutions = mat_run(U, Q)
        return IppeCf._ippe_to_cf(solutions)

    @staticmethod
    def _cf_to_ippe(U_cf, Q_cf):
        modelDims = U_cf.shape[0]
        U_t = np.zeros_like(U_cf, dtype=float)
        Q_t = np.zeros_like(Q_cf, dtype=float)
        for i in range(modelDims):
            U_t[i] = IppeCf._rotate_vector_to_ippe(U_cf[i])
            Q_t[i] = np.array((-Q_cf[i][0], -Q_cf[i][1]))

        U = np.transpose(U_t)
        Q = np.transpose(Q_t)

        return U, Q

    @staticmethod
    def _ippe_to_cf(solutions):
        result = [
            IppeCf.Solution(
                IppeCf._rotate_rot_mat_to_cf(solutions['R1']),
                IppeCf._rotate_vector_to_cf(solutions['t1']),
                solutions['reprojError1']
            ),
            IppeCf.Solution(
                IppeCf._rotate_rot_mat_to_cf(solutions['R2']),
                IppeCf._rotate_vector_to_cf(solutions['t2']),
                solutions['reprojError2']
            )
        ]

        return result

    @staticmethod
    def _rotate_vector_to_ippe(v):
        return np.dot(IppeCf._R_cf_to_ippe, v)

    def _rotate_vector_to_cf(v):
        return np.dot(IppeCf._R_ippe_to_cf, v)

    @staticmethod
    def _rotate_rot_mat_to_cf(R):
        return np.dot(IppeCf._R_ippe_to_cf, np.dot(R, IppeCf._R_cf_to_ippe))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2023 Bitcraze AB
#
#  Crazyflie Python Library
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Example script that makes the Crazyflie follow a wall

This examples uses the Flow and Multi-ranger decks to measure distances
in all directions and do wall following. Straight walls with corners
are advised to have in the test environment.
This is a python port of c-based app layer example from the Crazyflie-firmware
found here https://github.com/bitcraze/crazyflie-firmware/tree/master/examples/
demos/app_wall_following_demo

For the example to run the following hardware is needed:
 * Crazyflie 2.0
 * Crazyradio PA
 * Flow deck
 * Multiranger deck
"""
import logging
import time
from math import degrees
from math import radians

from wall_following.wall_following import WallFollowing

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.crazyflie.syncLogger import SyncLogger
from cflib.positioning.motion_commander import MotionCommander
from cflib.utils import uri_helper
from cflib.utils.multiranger import Multiranger

URI = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')


def handle_range_measurement(range):
    if range is None:
        range = 999
    return range


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    # Only output errors from the logging framework
    logging.basicConfig(level=logging.ERROR)

    keep_flying = True

    wall_following = WallFollowing(
        angle_value_buffer=0.1, reference_distance_from_wall=0.3,
        max_forward_speed=0.3, init_state=WallFollowing.StateWallFollowing.FORWARD)

    # Setup logging to get the yaw data
    lg_stab = LogConfig(name='Stabilizer', period_in_ms=100)
    lg_stab.add_variable('stabilizer.yaw', 'float')

    cf = Crazyflie(rw_cache='./cache')
    with SyncCrazyflie(URI, cf=cf) as scf:
        # Arm the Crazyflie
        scf.cf.platform.send_arming_request(True)
        time.sleep(1.0)

        with MotionCommander(scf) as motion_commander:
            with Multiranger(scf) as multiranger:
                with SyncLogger(scf, lg_stab) as logger:

                    while keep_flying:

                        # initialize variables
                        velocity_x = 0.0
                        velocity_y = 0.0
                        yaw_rate = 0.0
                        state_wf = WallFollowing.StateWallFollowing.HOVER

                        # Get Yaw
                        log_entry = logger.next()
                        data = log_entry[1]
                        actual_yaw = data['stabilizer.yaw']
                        actual_yaw_rad = radians(actual_yaw)

                        # get front range in meters
                        front_range = handle_range_measurement(multiranger.front)
                        top_range = handle_range_measurement(multiranger.up)
                        left_range = handle_range_measurement(multiranger.left)
                        right_range = handle_range_measurement(multiranger.right)

                        # choose here the direction that you want the wall following to turn to
                        wall_following_direction = WallFollowing.WallFollowingDirection.RIGHT
                        side_range = left_range

                        # get velocity commands and current state from wall following state machine
                        velocity_x, velocity_y, yaw_rate, state_wf = wall_following.wall_follower(
                            front_range, side_range, actual_yaw_rad, wall_following_direction, time.time())

                        print('velocity_x', velocity_x, 'velocity_y', velocity_y,
                              'yaw_rate', yaw_rate, 'state_wf', state_wf)

                        # convert yaw_rate from rad to deg
                        yaw_rate_deg = degrees(yaw_rate)

                        motion_commander.start_linear_motion(
                            velocity_x, velocity_y, 0, rate_yaw=yaw_rate_deg)

                        # if top_range is activated, stop the demo
                        if top_range < 0.2:
                            keep_flying = False



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2024 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import yaml

from cflib.crazyflie.param import PersistentParamState


class ParamFileManager():
    """Reads and writes parameter configurations from file"""
    TYPE_ID = 'type'
    TYPE = 'persistent_param_state'
    VERSION_ID = 'version'
    VERSION = '1'
    PARAMS_ID = 'params'

    @staticmethod
    def write(file_name, params={}):
        file = open(file_name, 'w')
        with file:
            file_params = {}
            for id, param in params.items():
                assert isinstance(param, PersistentParamState)
                if isinstance(param, PersistentParamState):
                    file_params[id] = {'is_stored': param.is_stored,
                                       'default_value': param.default_value, 'stored_value': param.stored_value}

            data = {
                ParamFileManager.TYPE_ID: ParamFileManager.TYPE,
                ParamFileManager.VERSION_ID: ParamFileManager.VERSION,
                ParamFileManager.PARAMS_ID: file_params
            }

            yaml.dump(data, file)

    @staticmethod
    def read(file_name):
        file = open(file_name, 'r')
        with file:
            data = None
            try:
                data = yaml.safe_load(file)
            except yaml.YAMLError as exc:
                print(exc)

            if ParamFileManager.TYPE_ID not in data:
                raise Exception('Type field missing')

            if data[ParamFileManager.TYPE_ID] != ParamFileManager.TYPE:
                raise Exception('Unsupported file type')

            if ParamFileManager.VERSION_ID not in data:
                raise Exception('Version field missing')

            if data[ParamFileManager.VERSION_ID] != ParamFileManager.VERSION:
                raise Exception('Unsupported file version')

            def get_data(input_data):
                persistent_params = {}
                for id, param in input_data.items():
                    persistent_params[id] = PersistentParamState(
                        param['is_stored'], param['default_value'], param['stored_value'])
                return persistent_params

            if ParamFileManager.PARAMS_ID in data:
                return get_data(data[ParamFileManager.PARAMS_ID])
            else:
                return {}
